package dev.pfaff.altargui;

import dev.pfaff.altargui.debug.events.MountEvent;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.render.CanvasImpl;
import dev.pfaff.altargui.internal.render.ScissorStack;
import it.unimi.dsi.fastutil.shorts.ShortArrayList;

import java.util.List;

import static dev.pfaff.altargui.GuiGlobals.LOGGER;
import static dev.pfaff.altargui.GuiGlobals.profiler;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.util.FormatProcessor.FMT;

/**
 * Responsible for including children for input events and painting.
 */
public final class MountContext {
	private static final List<String> TRACE_TAGS = List.of(DEBUG, "MOUNT_OP");

	private final CanvasImpl canvas;
	private final WidgetInstanceList orderedBuffer;
	private final ShortArrayList treeBuffer;

	private final ScissorStack scissorStack = new ScissorStack();
	private float scissorTopMinX = Float.NaN;
	private float scissorTopMinY = Float.NaN;
	private float scissorTopMaxX = Float.NaN;
	private float scissorTopMaxY = Float.NaN;

	long time;
	boolean generation;
	private short treeDepth = 0;

	MountContext(CanvasImpl canvas, WidgetInstanceList orderedBuffer, ShortArrayList treeBuffer) {
		this.canvas = canvas;
		this.orderedBuffer = orderedBuffer;
		this.treeBuffer = treeBuffer;
	}

	public void pushScissor(float x, float y, float width, float height) {
		LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.pushScissor(\{x}, \{y}, \{width}, \{height})", TRACE_TAGS);
		this.canvas.internalPushScissor(x, y, width, height);
		var maxX = x + width;
		var maxY = y + height;
		this.scissorTopMinX = x;
		this.scissorTopMinY = y;
		this.scissorTopMaxX = maxX;
		this.scissorTopMaxY = maxY;
		this.scissorStack.push(x, y, maxX, maxY);
	}

	public void pushScissor(WidgetInstance<?> element) {
		pushScissor(element.clippedX(),
					element.clippedY(),
					element.clippedWidth(),
					element.clippedHeight());
	}

	public void popScissor() {
		LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.popScissor()", TRACE_TAGS);
		this.canvas.internalPopScissor();
		var scissorStack = this.scissorStack;
		scissorStack.pop();
		if (scissorStack.isEmpty()) {
			this.scissorTopMinX = Float.NaN;
			this.scissorTopMinY = Float.NaN;
			this.scissorTopMaxX = Float.NaN;
			this.scissorTopMaxY = Float.NaN;
		} else {
			this.scissorTopMinX = scissorStack.topMinX();
			this.scissorTopMinY = scissorStack.topMinY();
			this.scissorTopMaxX = scissorStack.topMaxX();
			this.scissorTopMaxY = scissorStack.topMaxY();
		}
	}

	private boolean shouldCull(WidgetInstance<?> element) {
		float x = element.allocatedX();
		float y = element.allocatedY();
		float width = element.computedWidth();
		float height = element.computedHeight();
		var scissorMinX = this.scissorTopMinX;
		var scissorMinY = this.scissorTopMinY;
		var scissorMaxX = this.scissorTopMaxX;
		var scissorMaxY = this.scissorTopMaxY;
		// comparison is false if either side is NaN
		// OR them all to find out if any are outside
		var shouldCull = scissorMinX > x + width |
						 scissorMinY > y + height |
						 x > scissorMaxX |
						 y > scissorMaxY;
//			element.lastShouldPaint = !shouldCull;
		return shouldCull;
	}

	private short mountCommon0(WidgetInstance<?> element) {
		this.orderedBuffer.add(element);
		short depth = this.treeDepth++;
		this.treeBuffer.add(depth);
		return depth;
	}

	public void mountImpl(WidgetInstance<?> element) {
		assert element != null : "Attempted to mount a null element";

		profiler().pushEvent(new MountEvent(element.key()));
		try {
			LOGGER.log(() -> STR."\{" ".repeat(Short.toUnsignedInt(this.treeDepth))}MountContext.mount(\{element.toStringShort()})", TRACE_TAGS);

			element.generation = this.generation;

//				if (shouldCull(element)) {
//					// FIXME: because this skips calling element.mount(), we can end up with elements being disposed
//					//  that are still references in other elements. That wouldn't be an issue, except that those if
//					//  those elements become visible again, they will no longer be remembered by the GuiContext, and
//					//  will never be disposed again.
////					LOGGER.log(Level.INFO, "culling");
//					return;
//				}

			if (element.painter != null) {
				profiler().push("paint");
				element.painter.paint(this.time, this.canvas);
				profiler().pop();
			}

			var depth = this.mountCommon0(element);

			if (Config.CHECKS && (!element.hasStencilLaidOut() || !element.hasTraceLaidOut() || !element.hasScissorLaidOut())) {
				throw new IllegalStateException("Attempted to mount an element that has never been laid out: " + this);
			}
			if (Config.CHECKS && (element.needsTraceLayout.get() || element.needsScissorLayout.get())) {
				throw new IllegalStateException("Attempted to mount an element that needs re-layout: " + this);
			}
			element.computeMount(this);

			assert this.treeDepth == depth + 1;
			this.treeDepth = depth;

			if (element.postPainter != null) {
				profiler().push("paint");
				element.postPainter.paintPost(this.time, this.canvas);
				profiler().pop();
			}

			// null-checking here with a cache field instead of instanceof-checking for better performance.
			if (element.debugLayer != null) {
				assert element.debugRect != null;
				profiler().push("paint");
				if (this.canvas.isDebugLayerEnabled(element.debugLayer)) {
					element.debugRect.updatePositionAndSize(element.allocatedX(),
							element.allocatedY(),
							element.computedWidth(),
							element.computedHeight());
					this.canvas.upload(element.debugRect);
				}
				profiler().pop();
			}
		} catch (Throwable e) {
			throw GuiException.wrap(FMT."Mounting \{element.toStringShort()}\n\{element}", e);
		} finally {
			profiler().popEvent();
		}
	}

	/**
	 * @deprecated {@link WidgetInstanceList} offers improved performance over generic {@link List}.
	 */
	@Deprecated
	public void mountList(List<? extends WidgetInstance<?>> elements) {
		elements.forEach(element -> element.mount(this));
	}

	public void mountList(WidgetInstanceList elements) {
		mountArray(elements.elements(), elements.size());
	}

	public void mountArray(WidgetInstance<?>[] elements) {
		mountArray(elements, elements.length);
	}

	public void mountArray(WidgetInstance<?>[] elements, int count) {
		if (count > elements.length) throw new AssertionError();
		for (int i = 0; i < count; i++) {
			elements[i].mount(this);
		}
	}
}
