package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.async.Task;
import dev.pfaff.altargui.async.TaskExecutor;
import dev.pfaff.altargui.function.FallibleRunnable;
import dev.pfaff.altargui.host.ContextForker;
import dev.pfaff.altargui.host.ForkedContext;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.altargui.util.TaskException;
import dev.pfaff.glass.opengl.GL;
import dev.pfaff.log4truth.Logger;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.LockSupport;
import java.util.stream.IntStream;

import static dev.pfaff.altargui.GuiGlobals.HOST;
import static dev.pfaff.altargui.async.VirtualThreadTask.spawnVirtualTask;
import static dev.pfaff.glass.Config.getEnvInt;
import static dev.pfaff.log4truth.StandardTags.ERROR;

/**
 * A pool of OpenGL worker threads, each with their own context with sharing enabled with the main context.
 */
public abstract sealed class GLContextPool implements TaskExecutor {
	public static ScopedValue<GL> POOL_ACQUIRED_GL = ScopedValue.newInstance();

	/**
	 * The global context pool.
	 */
	public static final GLContextPool GL_CONTEXT_POOL;

	static {
		var forker = HOST.contextForker();
		if (forker != null) {
			GL_CONTEXT_POOL = new PerCore(forker);
		} else {
			GL_CONTEXT_POOL = new OnMainThread();
		}
	}

	/**
	 * To ensure optimal scheduling, each independent task should request a new executor with a call to this method.
	 */
	public abstract TaskExecutor acquireExecutor();

	@Override
	public <V> Task<V> submit(Callable<V> callable) {
		return acquireExecutor().submit(callable);
	}

	@Override
	public Task<Void> submit(Runnable runnable) {
		return acquireExecutor().submit(runnable);
	}

	@Override
	public Task<Void> submitFallible(FallibleRunnable runnable) {
		return acquireExecutor().submitFallible(runnable);
	}

	@Override
	public <V> V execute(Callable<V> callable) {
		return acquireExecutor().execute(callable);
	}

	@Override
	public void execute(Runnable runnable) {
		acquireExecutor().execute(runnable);
	}

	@Override
	public void executeFallible(FallibleRunnable runnable) {
		acquireExecutor().executeFallible(runnable);
	}

	private static final class OnMainThread extends GLContextPool {
		private static final TaskExecutor EXECUTOR = new TaskExecutor() {
			@Override
			public <V> Task<V> submit(Callable<V> callable) {
				return GLThread.executor().submit(() -> {
					return ScopedValue.callWhere(POOL_ACQUIRED_GL, GuiGlobals.GL, callable::call);
				});
			}
		};

		@Override
		public TaskExecutor acquireExecutor() {
			return EXECUTOR;
		}
	}

	private static final class PerCore extends GLContextPool {
		private static final int GL_CONTEXT_POOL_SIZE = getEnvInt(
			"ALTAR_GUI_GL_CONTEXT_POOL_SIZE",
			// leave one core for the main thread.
			Math.max(Runtime.getRuntime().availableProcessors() - 1, 1)
		);

		private final List<ExecutorImpl> threadTasks;

		public PerCore(ContextForker forker) {
			this.threadTasks = IntStream.range(0, GL_CONTEXT_POOL_SIZE)
					.mapToObj(i -> {
						var executor = new ExecutorImpl(forker, i);
						executor.thread.start();
						return executor;
					})
					.toList();
		}

		@Override
		public TaskExecutor acquireExecutor() {
			return this.threadTasks.stream()
								   .unordered()
								   .min((a, b) -> {
									   // prefer less busy
									   return Integer.compare(a.thread.queue.size(), b.thread.queue.size());
								   })
								   .get();
		}

		private static final class ThreadImpl extends Thread {
			private final ContextForker forker;
			//@Stable
			private ForkedContext context;
			private final ConcurrentLinkedQueue<Runnable> queue = new ConcurrentLinkedQueue<>();

			public ThreadImpl(ContextForker forker, int index) {
				super(STR."gl-worker-\{index}");
				this.forker = forker;
				setDaemon(true);
			}

			@Override
			public void run() {
				try (var _ = this.context = forker.forkContext()) {
					for (; ; ) {
						Runnable task;
						while ((task = queue.poll()) != null) {
							try {
								task.run();
							} catch (Throwable e) {
								Logger.log(() -> "Caught exception thrown by scheduled task. That should not happen!", e, List.of(ERROR));
							}
						}

						// park or continue if unparked
						LockSupport.park();
						if (Thread.interrupted()) {
							throw new RuntimeException(new InterruptedException());
						}
					}
				}
			}
		}

		private static class ExecutorImpl implements TaskExecutor {
			private final ThreadImpl thread;

			public ExecutorImpl(ContextForker forker, int index) {
				this.thread = new ThreadImpl(forker, index);
			}

			@Override
			public <V> Task<V> submit(Callable<V> task) {
				return spawnVirtualTask(() -> {
					return ScopedValue.where(POOL_ACQUIRED_GL, thread.context.gl())
									  .call(task::call);
				}, task1 -> {
					thread.queue.add(task1);
					LockSupport.unpark(thread);
				});
			}
		}
	}
}
