package dev.pfaff.altargui.render;

import dev.pfaff.glass.Interop;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.enums.GLDebugSeverity;
import dev.pfaff.glass.opengl.enums.GLDebugSource;
import dev.pfaff.glass.opengl.enums.GLDebugType;
import dev.pfaff.log4truth.Logger;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.FunctionDescriptor;
import java.lang.foreign.Linker;
import java.lang.foreign.MemorySegment;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.util.List;

import static dev.pfaff.altargui.internal.InternalGlobals.GL_FEATURES;
import static dev.pfaff.glass.Interop.ADDRESS_INT;
import static java.lang.foreign.ValueLayout.ADDRESS;
import static java.lang.foreign.ValueLayout.JAVA_INT;
import static java.lang.invoke.MethodHandles.filterArguments;
import static java.lang.invoke.MethodType.methodType;
import static java.util.FormatProcessor.FMT;

public final class GLDebugCallback implements AutoCloseable {
	private static final class Lazy {
		private static final MethodHandle MH_handleMessage;
		private static FunctionDescriptor FD_glDebugMessageCallback;

		static {
			MethodHandle mh;
			try {
				mh = MethodHandles.lookup().findStatic(GLDebugCallback.Lazy.class,
													   "handleMessage",
													   methodType(void.class,
																  int.class,
																  int.class,
																  int.class,
																  int.class,
																  long.class,
																  MemorySegment.class,
																  MemorySegment.class));
			} catch (NoSuchMethodException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
			MH_handleMessage = filterArguments(mh, 4, Interop.MH_cast_usize_to_long);

			FD_glDebugMessageCallback = FunctionDescriptor.ofVoid(JAVA_INT,
																  JAVA_INT,
																  JAVA_INT,
																  JAVA_INT,
																  ADDRESS_INT,
																  ADDRESS,
																  ADDRESS);
		}

		private static final List<List<String>> TAGS_MATRIX;

		static {
			assert GLDebugSource.DONT_CARE.ordinal() == 0;
			assert GLDebugType.DONT_CARE.ordinal() == 0;
			assert GLDebugSeverity.DONT_CARE.ordinal() == 0;
			TAGS_MATRIX = GLDebugSource.VALUES.stream()
											  .filter(source -> source != GLDebugSource.DONT_CARE)
											  .map(source -> "GL" + source)
											  .flatMap(source -> GLDebugType.VALUES.stream()
																				   .filter(type -> type != GLDebugType.DONT_CARE)
																				   .map(type -> "GL" + type)
																				   .flatMap(type -> GLDebugSeverity.VALUES.stream()
																														  .filter(severity -> severity != GLDebugSeverity.DONT_CARE)
																														  .map(severity -> List.of(source, type, "GL" + severity))))
											  .toList();
		}

		private static void handleMessage(GLDebugSource source,
										  GLDebugType type,
										  int id,
										  GLDebugSeverity severity,
										  long length,
										  ConstPointer<Byte> message,
										  @Nullable ConstPointer<Void> userParam) {
			assert userParam == null;
			var tagsMatrixIndex = (source.ordinal() - 1) * (GLDebugType.VALUES.size() - 1) * (GLDebugSeverity.VALUES.size() - 1)
								  + (type.ordinal() - 1) * (GLDebugSeverity.VALUES.size() - 1)
								  + (severity.ordinal() - 1);
			Logger.log(() -> {
				return FMT."0x%08x\{id}: \{message.ms().reinterpret(length + 1).getString(0)}";
			}, Lazy.TAGS_MATRIX.get(tagsMatrixIndex));
		}

		private static void handleMessage(int source,
										  int type,
										  int id,
										  int severity,
										  long length,
										  MemorySegment message,
										  MemorySegment userParam) {
			handleMessage(GLDebugSource.fromGLEnum(source),
						  GLDebugType.fromGLEnum(type),
						  id,
						  GLDebugSeverity.fromGLEnum(severity),
						  length,
						  new ConstPointer<>(message),
						  ConstPointer.newNullable(userParam));
		}
	}

	private final Arena arena;

	private GLDebugCallback(Arena arena) {
		this.arena = arena;
	}

	public static GLDebugCallback register() {
		var arena = Arena.ofShared();
		try {
			var khrDebug = GL_FEATURES.getKHR_Debug();
			if (khrDebug != null) {
				var stub = Linker.nativeLinker().upcallStub(Lazy.MH_handleMessage, Lazy.FD_glDebugMessageCallback, arena)
					.reinterpret(arena, _ -> {
						GL_FEATURES.getKHR_Debug().glDebugMessageCallback(null, null);
					});
				khrDebug.glDebugMessageCallback(new ConstPointer<>(stub), null);
			}
			return new GLDebugCallback(arena);
		} catch (Throwable t) {
			arena.close();
			throw t;
		}
	}

	@Override
	public void close() throws Exception {
		arena.close();
	}
}
