package dev.pfaff.altargui.render;

import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.enums.*;
import org.jetbrains.annotations.NotNull;

import java.lang.foreign.MemorySegment;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.internal.Config.GL_CHECKS;

public final class GLUtil {
	public static @NotNull MemorySegment glMapBuffer(GLBufferTarget target, GLBufferAccess access, long length) {
		var addr = GL.glMapBuffer(target, access);
		if (addr == null) throw new NullPointerException();
		var buffer = addr.ms().reinterpret(length);
		return switch (access) {
			case READ_ONLY -> buffer.asReadOnly();
			case WRITE_ONLY, READ_WRITE -> buffer;
		};
	}

	/**
	 * @param access bitset of {@code GL_MAP_*} bits.
	 */
	public static @NotNull MemorySegment glMapBufferRange(GLBufferTarget target, long offset, long length,
														  int access) {
		var addr = GL.glMapBufferRange(target, offset, length, access);
		if (addr == null) throw new NullPointerException();
		var buffer = addr.ms().reinterpret(length);
		return (access & GLMapBufferAccessMask.MAP_WRITE_BIT.glEnum()) == 0 ? buffer.asReadOnly() : buffer;
	}

	public static void glUnmapBuffer(GLBufferTarget target) {
		GL.glUnmapBuffer(target);
	}

	public static void glBufferData(GLBufferTarget target, long size, GLBufferUsage usage) {
		GL.glBufferData(target, size, null, usage);
	}

	public static void glBufferData(GLBufferTarget target, @NotNull MemorySegment data, GLBufferUsage usage) {
		GL.glBufferData(target, data.byteSize(), ConstPointer.newNullable(data), usage);
	}

	public static void glBufferStorage(GLBufferStorageTarget target, int size, int flags) {
		GL.glBufferStorage(target, size, null, flags);
	}

	public static void glBufferSubData(GLBufferTarget target, long offset, @NotNull MemorySegment data) {
		GL.glBufferSubData(target, offset, data.byteSize(), ConstPointer.newNullable(data));
	}

	public static void glCopyBufferSubData(GLCopyBufferSubDataTarget readTarget, GLCopyBufferSubDataTarget writeTarget
			, long readOffset, long writeOffset, long length) {
		GL.glCopyBufferSubData(readTarget, writeTarget, readOffset, writeOffset, length);
	}

	public static void glVertexAttribDivisor(int location, int divisor) {
		GL.glVertexAttribDivisor(location, divisor);
	}

	public static void glBindVertexArray(int id) {
		GL.glBindVertexArray(id);
	}

	public static void glxUnbindVertexArray() {
		GL.glBindVertexArray(0);
	}

	public static int glGenBuffers() {
		return GL.glGenBuffer();
	}

	public static void glDeleteBuffers(int id) {
		GL.glDeleteBuffer(id);
	}

	public static int glGetAttribLocation(int program, String name) {
		return GL.glGetAttribLocation(program, name);
	}

	public static void glEnableVertexAttribArray(int id) {
		GL.glEnableVertexAttribArray(id);
	}

	/**
	 * @since OpenGL 2.0
	 */
	public static void glVertexAttribPointer(int id, int size, GLVertexAttribPointerType type, boolean normalized,
											 int stride, long pointer) {
		GL.glVertexAttribPointer(id,
				size,
				type,
				normalized,
				stride,
				ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	/**
	 * @since OpenGL 3.0
	 */
	public static void glVertexAttribIPointer(int id, int size, GLVertexAttribIType type, int stride, long pointer) {
		GL.glVertexAttribIPointer(id, size, type, stride, ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	/**
	 * @since OpenGL 4.1
	 */
	public static void glVertexAttribLPointer(int id, int size, GLVertexAttribLType type, int stride, long pointer) {
		GL.glVertexAttribLPointer(id, size, type, stride, ConstPointer.newNullable(MemorySegment.ofAddress(pointer)));
	}

	public static void checkBoundBuffer(GLGetPName pname, int id) {
		if (GL_CHECKS) {
			int boundId = GL.glGetInteger(pname);
			if (boundId != id) {
				throw new IllegalStateException(STR."Attempted to upload while not bound: expected \{id} found, \{boundId}");
			}
		}
	}

	/**
	 * @see <a href="https://docs.gl/gl4/glTexSubImage3D">docs.gl</a>
	 * @since OpenGL 2.0
	 */
	public static void glTexSubImage3D(GLTextureTarget target, int level, int xOffset, int yOffset, int zOffset,
									   int width, int height, int depth, GLPixelFormat format, GLPixelType type,
									   MemorySegment pixels) {
		if (GL_CHECKS) {
			var length = (long) width * height * depth;
			if (pixels.byteSize() < length) {
				throw new IndexOutOfBoundsException(STR."Dimensions \{width},\{height},\{depth} (length of \{length}) overflow buffer of size \{pixels.byteSize()}");
			}
		}
		GL.glTexSubImage3D(target,
				level,
				xOffset,
				yOffset,
				zOffset,
				width,
				height,
				depth,
				format,
				type,
				new ConstPointer<>(pixels));
	}
}
