package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.libimage.ColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.glass.opengl.enums.GLPixelFormat;
import dev.pfaff.glass.opengl.enums.GLPixelType;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;

import static java.lang.foreign.ValueLayout.JAVA_BYTE;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public record ImageSrc(MemorySegment pixels, GLPixelFormat format, GLPixelType type) {
	public static ImageSrc create(ImageData image) {
		return switch (image.colorType()) {
			case ColorType.Rgba8 -> new ImageSrc(swizzleRgba8ToBgra8(image.pixels()), GLPixelFormat.BGRA, GLPixelType.UNSIGNED_INT_8_8_8_8_REV);
			case ColorType.Rgb8 -> new ImageSrc(swizzleRgb8ToBgra8(image.pixels()), GLPixelFormat.BGRA, GLPixelType.UNSIGNED_INT_8_8_8_8_REV);
			default -> new ImageSrc(image.pixels(), image.colorType().glPixelFormat(), image.colorType().glChannelType());
		};
	}

	private static MemorySegment swizzleRgba8ToBgra8(MemorySegment pixels) {
		if (pixels.byteSize() % 4 != 0) {
			throw new IllegalArgumentException("Invalid RGBA8 pixels");
		}
		var dst = Arena.ofAuto().allocate(pixels.byteSize(), 16);
		for (long i = 0; i < pixels.byteSize(); i += 4) {
			var pixel = pixels.get(JAVA_INT, i);
			// RGBA bytes -> ABGR int
			// ABGR int rotl 8 -> BGRA int
			// BGRA int rev -> ARGB int
			// Integer.reverseBytes(Integer.rotateLeft(pixel, 8))
			dst.set(JAVA_INT, i, (pixel & 0xff_00_ff_00) | ((pixel & 0xff) << 16) | ((pixel & 0xff_00_00) >>> 16));
		}
		return dst;
	}

	private static MemorySegment swizzleRgb8ToBgra8(MemorySegment pixels) {
		if (pixels.byteSize() % 3 != 0) {
			throw new IllegalArgumentException("Invalid RGB8 pixels");
		}
		var dst = Arena.ofAuto().allocate(pixels.byteSize() / 3 * 4, 16);
		for (long i = 0, j = 0; i < pixels.byteSize(); i += 3, j += 4) {
			var r = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i));
			var g = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i + 1));
			var b = Byte.toUnsignedInt(pixels.get(JAVA_BYTE, i + 2));
			dst.set(JAVA_INT, j, 0xff_00_00_00 | (r << 16) | (g << 8) | b);
		}
		return dst;
	}
}
