package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.async.NonCancellableTask;
import dev.pfaff.altargui.async.Task;
import dev.pfaff.altargui.async.VirtualThreadTask;
import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.io.IOUtil;
import dev.pfaff.altargui.libimage.EncoderProfile;
import dev.pfaff.altargui.libimage.ExtendedColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.libimage.ImageFormat;
import dev.pfaff.altargui.libimage.LibImage;
import dev.pfaff.altargui.util.Bits;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.glass.opengl.enums.GLGetPName;
import dev.pfaff.log4truth.NamedLogger;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import org.jspecify.annotations.Nullable;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.ref.SoftReference;
import java.nio.file.Path;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.Semaphore;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.async.Task.completedTask;
import static dev.pfaff.altargui.async.VirtualThreadTask.spawnVirtualTask;
import static dev.pfaff.altargui.internal.Config.TEXTURE_ATLAS_ARRAY_IMPL;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.ERROR;

/**
 * <p>
 * <code>
 * Arrays: | 1  | 2  | 3  |<br/>
 * Active: | D  | U* | U  |
 * </code>
 * <p>
 * States:
 * <ol>
 *     <li><strong>D</strong> clean for drawing</li>
 *     <li><strong>U</strong> accepting uploads</li>
 *     <li><strong>U*</strong> finishing upload</li>
 * </ol>
 * <p>
 * When a buffer with state {@code U*} is finished, it transitions to state {@code D}, the previous buffer
 * transitions to state {@code U}, and the next buffer transitions to state {@code U*}. Thanks to this model, the
 * states of all three buffers may be represented using a single 2-bit index.
 * <p>
 *
 * <code>
 * |---|---|---|<br/>
 * | A | B | C |<br/>
 * |---|---|---|
 * </code>
 * <p>
 * Initial state:
 * <ul>
 *     <li><strong>A)</strong> Drawing</li>
 *     <li><strong>B)</strong> Finishing (nothing)</li>
 *     <li><strong>C)</strong> Uploading</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job submission to C:
 * <ul>
 *     <li><strong>A)</strong> Uploading</li>
 *     <li><strong>B)</strong> Drawing</li>
 *     <li><strong>C)</strong> Finishing (at least 1 job)</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job completion in C:
 * <ul>
 *     <li><strong>A)</strong> Finishing (possibly nothing)</li>
 *     <li><strong>B)</strong> Uploading</li>
 *     <li><strong>C)</strong> Drawing</li>
 * </ul>
 *
 * <p>
 * Next state, triggered on job completion in A (if any) and job submission to B:
 * <ul>
 *     <li><strong>A)</strong> Drawing</li>
 *     <li><strong>B)</strong> Finishing (possibly nothing)</li>
 *     <li><strong>C)</strong> Uploading</li>
 * </ul>
 */
public final class TextureAllocator implements AutoCloseable {
	private static final NamedLogger LOGGER = new NamedLogger("TexAlloc");

	public static final int MAX_TEXTURE_SIZE;

	static {
		// cannot use lambdas across threads in static initializer
		MAX_TEXTURE_SIZE = GLThread.executor().execute(new Callable<Integer>() {
			@Override
			public Integer call() {
				return Math.min(GL.glGetInteger(GLGetPName.MAX_TEXTURE_SIZE), 256 * 256);
			}
		});
		if (MAX_TEXTURE_SIZE == 0) {
			throw new IllegalStateException("MAX_TEXTURE_SIZE must be greater than 0");
		}
	}

	/**
	 * The minimum allocation size in pixels. Herein we often measure in "units" rather than pixels for efficiency.
	 */
	public static final int UNIT = Math.divideExact(MAX_TEXTURE_SIZE, 256);
	public static final TextureAtlasScale UNIT_SCALE = TextureAtlasScale.fromAtlasSize(256, 256);

	static {
		if (Integer.bitCount(UNIT) != 1) {
			throw new UnsupportedOperationException(STR."Only power-of-two unit sizes are supported, but \{UNIT} was calculated");
		}
	}

	/**
	 * The size (in {@link #UNIT}s) of each layer in the allocator.
	 */
	private final int sizeUnits;
	private int layers;
	private final AbstractTextureArray textureArray = TEXTURE_ATLAS_ARRAY_IMPL.get();
	private MemorySegment allocations;
	private final Object2ObjectOpenHashMap<LoadableTexture, Object> cache =
			new Object2ObjectOpenHashMap<>();
	private final Semaphore lock = new Semaphore(1);

	/**
	 * @param size the width and height of each layer.
	 */
	public TextureAllocator(int size, int initialLayers) {
		if (size > MAX_TEXTURE_SIZE) {
			throw new IllegalArgumentException("Layer size too large");
		}
		this.sizeUnits = size / UNIT;
		if (size % UNIT != 0) throw new IllegalArgumentException(STR."Invalid layer size: \{size}");
		this.layers = initialLayers;
		// high alignment to encourage vectorization
		this.allocations = Arena.ofAuto().allocate(((long)this.sizeUnits * this.sizeUnits * this.layers) / Byte.SIZE, 16);

		reset();
	}

	public void reset() {
		var size = this.sizeUnits * UNIT;
		this.textureArray.bindAndAllocate(size, size, this.layers, TextureOptions.DEFAULT);;
	}

	public InfallibleAutoClosable boundForDraw() {
		return this.textureArray.bound();
	}

	/**
	 * <p>
	 * Gets a shader to the given texture. If not loaded, the texture will be loaded and a slot allocated for it.
	 * <p>
	 * This method must only be called from the main thread.
	 */
	public Task<TextureHandle> alloc(LoadableTexture key) {
		{
			var cacheValue = this.cache.get(key);
			// this inspection is wrong
			//noinspection ConstantValue
			if (cacheValue != null) {
				var handle = tryDerefCacheValue(cacheValue);
				if (handle != null) return handle;
			}
		}
		Task<TextureHandle> future = spawnVirtualTask(() -> {
			ImageData image;
			try {
				image = key.load();
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture load failed", t, List.of(ERROR));
				return GLThread.executor().execute(() -> {
					return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
				});
			}

			try {
				// allocate a new image in the texture array.
				var handle = this.allocNew(key, image);

				// this will block until well beyond this.cache.put(key, future) below
				GLThread.executor().execute(() -> {
					assert this.cache.get(key) instanceof VirtualThreadTask<?> : this.cache.get(key);
					this.cache.put(key, new SoftReference<>(handle));
				});
				return handle;
			} catch (Throwable t) {
				LOGGER.log(() -> "Texture allocation failed", t, List.of(ERROR));
				return GLThread.executor().execute(() -> {
					return Objects.requireNonNull(tryDerefCacheValueJoin(this.cache.get(SinglePixelLoadableTexture.WHITE)));
				});
			}
		});
		this.cache.put(key, future);
		return new NonCancellableTask<>(future);
	}

	private static @Nullable Task<TextureHandle> tryDerefCacheValue(Object cacheValue) {
		if (cacheValue instanceof SoftReference<?> ref) {
			var handle = (TextureHandle) ref.get();
			if (handle != null) return completedTask(handle);
		} else {
			//noinspection unchecked
			return (Task<TextureHandle>) cacheValue;
		}
		return null;
	}

	private static @Nullable TextureHandle tryDerefCacheValueJoin(Object cacheValue) throws InterruptedException {
		if (cacheValue instanceof SoftReference<?> ref) {
			return (TextureHandle) ref.get();
		} else {
			//noinspection unchecked
			return ((Task<TextureHandle>) cacheValue).join();
		}
	}

	private TextureHandle allocNew(LoadableTexture key, ImageData image) throws InterruptedException {
		int width = image.width();
		int height = image.height();

		// align size up to the next multiple of UNIT.
		// these are in units, not pixels
		int unitWidth = (width + (UNIT - 1)) / UNIT;
		int unitHeight = (height + (UNIT - 1)) / UNIT;
		assert unitWidth * UNIT >= width && unitHeight * UNIT >= height : STR."width and/or height was improperly aligned to unit size:\n\t\{width} > \{unitWidth * UNIT}\n\tor\n\t\{height} > \{unitHeight * UNIT}";

		int sizeUnits = this.sizeUnits;

		if (unitWidth > sizeUnits || unitHeight > sizeUnits) {
			throw new IllegalArgumentException(STR."Size of image exceeds size of allocator layers: \{width} || \{height} > \{this.sizeUnits * UNIT}");
		}

		LOGGER.log(() -> STR."Searching for a free region to allocate \{key} an image of \{width}x\{height}", List.of(DEBUG));

		DetailedHandle detailed;

		this.lock.acquire();
		try {
			// find a free region
			int allocationsSize = sizeUnits * sizeUnits * this.layers;
			int fromSearchUpr = allocationsSize - unitWidth - (unitHeight - 1) * sizeUnits;
			int layers = this.layers;
			int sizeUnitsSq = sizeUnits * sizeUnits;

			int xUpr = sizeUnits - unitHeight;

			int from = -1, iUpr = -1;

			search:
			for (int layerStart = 0; layerStart < layers * sizeUnitsSq; layerStart += sizeUnitsSq) {
				from = layerStart;
				int layerUpr = from + sizeUnitsSq;
				while (from <= fromSearchUpr) {
					assert from % sizeUnits == 0 :
							STR."not at the start of a row: from=\{from}";
					int rowUpr = from + xUpr;
					for (; from < rowUpr; from++) {
						iUpr = from + unitHeight * sizeUnits;
						assert iUpr <= layerUpr;

						if (isRegionFree(from, iUpr, unitWidth)) break search;
					}
					from += unitHeight;
				}
				iUpr = -1;
			}

			if (iUpr == -1) {
				throw new UnsupportedOperationException("TODO: realloc");
			}
			assert from != -1;

			detailed = this.makeHandle(from, iUpr, unitWidth, width, height);
		} finally {
			this.lock.release();
		}

		this.textureArray.bindAndUpload(ImageSrc.create(image), detailed.handle, detailed.u * UNIT, detailed.v * UNIT, width, height);
		return detailed.handle();
	}

	private DetailedHandle makeHandle(int from, int iUpr, int unitWidth, int textureWidth, int textureHeight) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// mark allocated
		for (int i = from; i < iUpr; i += sizeUnits) { // for each row
			// mark the region of each row allocated
			Bits.setBitsInRange(allocations, i, i + unitWidth);
		}

		var handle = MemorySegment.NULL.reinterpret(Arena.ofAuto(), _ -> {
			this.lock.acquireUninterruptibly();
			try {
				// mark free
				for (int i = from; i < iUpr; i += sizeUnits) {
					Bits.unsetBitsInRange(allocations, i, i + unitWidth);
				}
			} finally {
				this.lock.release();
			}
		});

		int layer = from / (sizeUnits * sizeUnits);
		int u = from % sizeUnits;
		int v = (from / sizeUnits) % sizeUnits;
		return new DetailedHandle(handle, layer, sizeUnits, u, v, textureWidth, textureHeight);
	}

	private boolean isRegionFree(int from, int iUpr, int unitWidth) {
		var allocations = this.allocations;
		int sizeUnits = this.sizeUnits;

		// check if all are free
		for (int i = from; i < iUpr; i += sizeUnits) {
			for (int x = 0; x < unitWidth; x++) {
				if (Bits.getBit(allocations, i + x)) return false;
			}
		}
		return true;
	}

	@Override
	public void close() {
		this.textureArray.close();
	}

	public void dump() {
		this.textureArray.dump();
		try {
			this.lock.acquireUninterruptibly();
			try {
				try (var arena = Arena.ofConfined()) {
					var height = Math.multiplyExact(this.sizeUnits, this.layers);
					var encoded = LibImage.encodeImage(this.allocations,
							this.sizeUnits,
							height,
							ExtendedColorType.L1,
							ImageFormat.Png,
							EncoderProfile.Fast,
							arena);
					IOUtil.write(Path.of("allocated-units.png"), encoded);
				}
			} finally {
				this.lock.release();
			}
		} catch (IOException e) {
			LOGGER.log(() -> "Failed to dump texture atlas", e, List.of(ERROR));
		}
	}

	/**
	 * @param u measured in {@link #UNIT}s
	 * @param v measured in {@link #UNIT}s
	 */
	record DetailedHandle(TextureHandle handle, int u, int v) {
		/**
		 * @param layer            index
		 * @param layerSizeInUnits the size of each layer, measured in {@link #UNIT}s
		 * @param u                measured in {@link #UNIT}s
		 * @param v                measured in {@link #UNIT}s
		 */
		public DetailedHandle(MemorySegment handle, int layer, int layerSizeInUnits, int u, int v, int textureWidth, int textureHeight) {
			assert Byte.toUnsignedInt((byte)u) == u;
			assert Byte.toUnsignedInt((byte)v) == v;
			this(new TextureHandle(handle,
					layer,
					(byte) u,
					(byte) v,
					TextureAtlasScale.fromAtlasSize(layerSizeInUnits * UNIT, layerSizeInUnits * UNIT),
					textureWidth,
					textureHeight), u, v);
		}
	}
}
