package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.io.IOUtil;
import dev.pfaff.altargui.libimage.EncoderProfile;
import dev.pfaff.altargui.libimage.ExtendedColorType;
import dev.pfaff.altargui.libimage.ImageFormat;
import dev.pfaff.altargui.libimage.LibImage;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.MutPointer;
import dev.pfaff.glass.opengl.enums.*;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.log4truth.NamedLogger;
import org.jspecify.annotations.Nullable;

import java.io.IOException;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.nio.file.Path;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicInteger;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLContextPool.GL_CONTEXT_POOL;
import static dev.pfaff.altargui.render.GLContextPool.POOL_ACQUIRED_GL;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.StandardTags.ERROR;
import static dev.pfaff.log4truth.StandardTags.WARN;
import static dev.pfaff.log4truth.format.Format.STANDARD_DURATION;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class TextureArray extends AbstractTextureArray {
	private static final NamedLogger LOGGER = NamedLogger.forCaller();

	public static final GLTextureTarget BIND_TARGET = GLTextureTarget.TEXTURE_2D;

	private static final AtomicInteger UPLOAD_ID = new AtomicInteger(0);

	public final int handle;
	private int width;
	private int height;
	private int layers;

	public TextureArray() {
		this.handle = GL.glGenTexture();
	}

	public int width() {
		return width;
	}

	public int height() {
		return height;
	}

	public int layers() {
		return layers;
	}

	@Override
	public InfallibleAutoClosable bound() {
		GL.glActiveTexture(GLTextureUnit.TEXTURE0);
		GL.glBindTexture(BIND_TARGET, handle);
		return Config.GL_ALWAYS_UNBIND ? () -> GL.glBindTexture(BIND_TARGET, 0) : () -> {};
	}

	/**
	 * Allocates the texture. It must already be {@link #bound}.
	 */
	@Override
	public void bindAndAllocate(int width, int height, int layerCount, TextureOptions options) {
		this.width = width;
		this.height = height;
		this.layers = layerCount;

		if (layerCount != 1) throw new AssertionError();

		try (var _ = Thread.pinToCarrierThread(); var _ = bound()) {
			GL.glTexStorage2D(BIND_TARGET, 1, GLSizedInternalFormat.RGBA8, width, height);
	//		try (var arena = Arena.ofConfined()) {
	//			var buf = arena.allocate(JAVA_INT, (long) width * height);
	//			buf.fill((byte) 0x00);
	//			GL.glTexImage2D(target(), 0, GLInternalFormat.RGBA8, width, height, 0, GLPixelFormat.RGBA, GLPixelType.UNSIGNED_INT_8_8_8_8_REV, new ConstPointer<>(buf));
	//		}

			GL.glTexParameteri(BIND_TARGET, GLTextureParameterName.TEXTURE_MIN_FILTER, options.minFilter().glEnum());
			GL.glTexParameteri(BIND_TARGET, GLTextureParameterName.TEXTURE_MAG_FILTER, options.magFilter().glEnum());
			if (options.clamp()) {
				GL.glTexParameteri(BIND_TARGET,
								   GLTextureParameterName.TEXTURE_WRAP_S,
								   GLTextureWrapMode.CLAMP_TO_EDGE.glEnum());
				GL.glTexParameteri(BIND_TARGET,
								   GLTextureParameterName.TEXTURE_WRAP_T,
								   GLTextureWrapMode.CLAMP_TO_EDGE.glEnum());
			}
		}
	}

	public void bindAndUpload(ImageSrc image, TextureHandle handle, int xOffset, int yOffset, int width, int height) {
		int id = UPLOAD_ID.getAndAdd(1);

		LOGGER.log(() -> STR."[Upload #\{id}] Uploading to \{handle} an image of \{width}x\{height} at \{xOffset},\{yOffset}", List.of(DEBUG));
		var completer = new CompletableFuture<@Nullable Void>();
		GL_CONTEXT_POOL.execute(() -> {
			try {
				var gl = POOL_ACQUIRED_GL.get();
				assert gl == GL;

				long start = System.nanoTime();

				var bufId = GL.glGenBuffer();
				try {
					//try (var _ = Thread.pinToCarrierThread()) {
					//	GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, bufId);
					//	try {
					//		GL.glBufferData(GLBufferTarget.PIXEL_UNPACK_BUFFER,
					//						src.pixels().byteSize(),
					//						new ConstPointer<>(src.pixels()),
					//						GLBufferUsage.STREAM_DRAW);
					//	} finally {
					//		GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, 0);
					//	}
					//}

					long startPinned = System.nanoTime();
					try (var _ = Thread.pinToCarrierThread();
						 //var _ = GLBufferBind.bindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, bufId);
						 var _ = bound()) {
						startUploading();
						//array.uploadFromPixelUnpackBuffer(handle.layer,
						//		detailed.u() * UNIT,
						//		detailed.v() * UNIT,
						//		width,
						//		height,
						//		src.format(),
						//		src.type());
						upload(handle.layer,
							   xOffset,
							   yOffset,
							   width,
							   height,
							   image.format(),
							   image.type(),
							   image.pixels());
					}

					long startSyncing = System.nanoTime();
					long elapsedPinned = startSyncing - startPinned;

					var syncId = GL.glFenceSync(GLSyncCondition.SYNC_GPU_COMMANDS_COMPLETE, 0);
					try {
						GL.glClientWaitSync(syncId, GLSyncObjectMask.SYNC_FLUSH_COMMANDS_BIT.glEnum(), Long.MAX_VALUE);
					} finally {
						GL.glDeleteSync(syncId);
					}

					long now = System.nanoTime();
					long elapsedSyncing = now - startSyncing;
					long elapsed = now - start;
					LOGGER.log(() -> STR."""
								[Upload #\{id}] Uploaded to textureArray=\{this.handle}
								  took \{STANDARD_DURATION.nanosToString(elapsed)} in total
								    \{STANDARD_DURATION.nanosToString(elapsedPinned)} pinned
								    \{STANDARD_DURATION.nanosToString(elapsedSyncing)} syncing""", List.of(DEBUG));
				} finally {
					GL.glDeleteBuffer(bufId);
				}
			} catch (Throwable e) {
				completer.completeExceptionally(e);
			}
		});
		completer.join();
		LOGGER.log(() -> STR."[Upload #\{id}] Uploaded image to \{handle}", List.of(DEBUG));
	}

	public void startUploading() {
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_ROW_LENGTH, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_SKIP_PIXELS, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_SKIP_ROWS, 0);
		GL.glPixelStorei(GLPixelStoreParameter.UNPACK_ALIGNMENT, 4);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void upload(int layer, int xOffset, int yOffset, int width, int height, GLPixelFormat format, GLPixelType type, MemorySegment pixels) {
		glTexSubImage2DChecks(layer, xOffset, yOffset, width, height, format, type);
		GL.glTexSubImage2D(BIND_TARGET,
						   0,
						   xOffset,
						   yOffset,
						   width,
						   height,
						   format,
						   type,
						   new ConstPointer<>(pixels));
	}

	private void glTexSubImage2DChecks(int layer, int xOffset, int yOffset, int width, int height, GLPixelFormat format, GLPixelType type) {
		if (0 > xOffset || 0 > yOffset || xOffset > this.width || yOffset > this.height) {
			throw new IllegalArgumentException(STR."Upload offset out of bounds: \{xOffset} !in [0, \{this.width}] || \{yOffset} !in [0, \{this.height}]");
		}
		var availWidth = this.width - xOffset;
		var availHeight = this.height - yOffset;
		if (0 > width || 0 > height || width > availWidth || height > availHeight) {
			throw new IllegalArgumentException(STR."Upload size out of bounds: \{width} !in [0, \{availWidth}] || \{height} !in [0, \{availHeight}]");
		}
		if (layer >= this.layers) {
			throw new IllegalArgumentException(STR."Upload layer out of bounds: \{layer} >= \{this.layers}");
		}

		if (format != GLPixelFormat.BGRA || type != GLPixelType.UNSIGNED_INT_8_8_8_8_REV) {
			Logger.log(() -> STR."Non-optimal texture upload! format=\{format}, type=\{type}", List.of(WARN));
		}

		if (layer != 0) throw new AssertionError();
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void uploadFromPixelUnpackBuffer(int layer,
											int xOffset, int yOffset,
											int width, int height,
											GLPixelFormat format,
											GLPixelType type) {
		glTexSubImage2DChecks(layer, xOffset, yOffset, width, height, format, type);
		GL.glTexSubImage2D(BIND_TARGET,
						   0,
						   xOffset,
						   yOffset,
						   width,
						   height,
						   format,
						   type,
						   null);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void upload(int layer, int width, int height, GLPixelFormat format, GLPixelType type, MemorySegment pixels) {
		this.upload(layer, 0, 0, width, height, format, type, pixels);
	}

	/**
	 * Remember to call {@link #startUploading()}.
	 */
	public void uploadFull(GLPixelFormat format, GLPixelType type, MemorySegment pixels) {
		long expected = (long) width * height * layers * 4;
		if (pixels.byteSize() != expected) {
			throw new IllegalArgumentException(STR."Pixel buffer size is wrong: \{pixels.byteSize()} != \{expected}");
		}

		GL.glTexSubImage2D(BIND_TARGET,
						   0,
						   0,
						   0,
						   width,
						   height,
						   format,
						   type,
						   new ConstPointer<>(pixels));
	}

	@Override
	public void close() {
		GL.glDeleteTexture(this.handle);
	}

	@Override
	public void dump() {
		try (var arena = Arena.ofShared()) {
			int width = this.width;
			int height = this.height * this.layers;
			var segment = arena.allocate(JAVA_INT, (long) width * height);
			GL_CONTEXT_POOL.execute(() -> {
				try (var _ = Thread.pinToCarrierThread(); var _ = bound()) {
					GL.glGetTexImage(TextureArray.BIND_TARGET,
									 0,
									 GLPixelFormat.RGBA,
									 GLPixelType.UNSIGNED_BYTE,
									 new MutPointer<>(segment));
				}
			});
			var encoded = LibImage.encodeImage(segment,
											   width,
											   height,
											   ExtendedColorType.Rgba8,
											   ImageFormat.Png,
											   EncoderProfile.Small,
											   arena);
			IOUtil.write(Path.of(STR."allocated-texture-\{handle}.png"), encoded);
		} catch (IOException e) {
			LOGGER.log(() -> "Failed to dump texture array", e, List.of(ERROR));
		}
	}
}
