package dev.pfaff.altargui.texture;

import dev.pfaff.altargui.function.InfallibleAutoClosable;
import dev.pfaff.glass.opengl.enums.GLSyncCondition;
import dev.pfaff.glass.opengl.enums.GLSyncObjectMask;
import dev.pfaff.log4truth.NamedLogger;
import org.jetbrains.annotations.NotNull;
import org.jspecify.annotations.Nullable;

import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Phaser;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLContextPool.GL_CONTEXT_POOL;
import static dev.pfaff.altargui.render.GLContextPool.POOL_ACQUIRED_GL;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static dev.pfaff.log4truth.format.Format.STANDARD_DURATION;

public final class TripleBufferedTextureArray extends AbstractTextureArray implements AutoCloseable {
	private static final NamedLogger LOGGER = NamedLogger.forCaller();
	public static final byte TBS_COUNT = 3;

	private static final AtomicInteger UPLOAD_ID = new AtomicInteger(0);

	private final AtomicReference<TBS> tbs = new AtomicReference<>(new TBS());
	private final TextureArray textureArrayA = new TextureArray();
	private final TextureArray textureArrayB = new TextureArray();
	private final TextureArray textureArrayC = new TextureArray();

	@Override
	public InfallibleAutoClosable bound() {
		var tbs = this.tbs.get();
		var nextTbs = tbs.pollNext();
		if (nextTbs != null) {
			tbs = rotateTBS(tbs, nextTbs);
		}
		return this.getArray(tbs.rotation).bound();
	}

	@Override
	public void bindAndAllocate(int width, int height, int layerCount, TextureOptions options) {
		for (byte i = 0; i < TBS_COUNT; i++) {
			var textureArray = this.getArray(i);
			try (var _ = textureArray.bound()) {
				textureArray.bindAndAllocate(width, height, layerCount, options);
			}
		}
	}

	public void bindAndUpload(ImageSrc image, TextureHandle handle, int xOffset, int yOffset, int width, int height) {
		int id = UPLOAD_ID.getAndAdd(1);

		LOGGER.log(() -> STR."[Upload #\{id}] Uploading to \{handle} an image of \{width}x\{height} at \{xOffset},\{yOffset}", List.of(DEBUG));
		var completer = new CompletableFuture<@Nullable Void>();
		GL_CONTEXT_POOL.execute(() -> {
			try {
				var gl = POOL_ACQUIRED_GL.get();
				assert gl == GL;

				var bufId = GL.glGenBuffer();
				try {
					//try (var _ = Thread.pinToCarrierThread()) {
					//	GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, bufId);
					//	try {
					//		GL.glBufferData(GLBufferTarget.PIXEL_UNPACK_BUFFER,
					//						src.pixels().byteSize(),
					//						new ConstPointer<>(src.pixels()),
					//						GLBufferUsage.STREAM_DRAW);
					//	} finally {
					//		GL.glBindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, 0);
					//	}
					//}

					TBS tbs;
					while (true) {
						tbs = this.tbs.get();
						if (tbs.initiating.activeCounter.addAndGet(1) > 0) break;
						// undo
						tbs.initiating.activeCounter.getAndAdd(-1);
					}
					for (int i = 0; i < TBS_COUNT;) {
						long start = System.nanoTime();
						tbs.initiating.dirty.set(true);

						var array = this.getArray((byte) ((byte) (tbs.rotation + 2) % TBS_COUNT));

						// upload
						try {
							long startPinned = System.nanoTime();
							try (var _ = Thread.pinToCarrierThread();
								 //var _ = GLBufferBind.bindBuffer(GLBufferTarget.PIXEL_UNPACK_BUFFER, bufId);
								 var _ = array.bound()) {
								array.startUploading();
								//array.uploadFromPixelUnpackBuffer(handle.layer,
								//		detailed.u() * UNIT,
								//		detailed.v() * UNIT,
								//		width,
								//		height,
								//		src.format(),
								//		src.type());
								array.upload(handle.layer,
											 xOffset,
											 yOffset,
											 width,
											 height,
											 image.format(),
											 image.type(),
											 image.pixels());
							}

							long startSyncing = System.nanoTime();
							long elapsedPinned = startSyncing - startPinned;

							var syncId = GL.glFenceSync(GLSyncCondition.SYNC_GPU_COMMANDS_COMPLETE, 0);
							try {
								GL.glClientWaitSync(syncId, GLSyncObjectMask.SYNC_FLUSH_COMMANDS_BIT.glEnum(), Long.MAX_VALUE);
							} finally {
								GL.glDeleteSync(syncId);
							}

							var oldTbs = tbs;
							long now = System.nanoTime();
							long elapsedSyncing = now - startSyncing;
							long elapsed = now - start;
							LOGGER.log(() -> STR."""
								[Upload #\{id}] Uploaded to uploadBuf=#\{tbBufIndex((byte) (oldTbs.rotation + 2))} (drawBuf=#\{tbBufIndex(oldTbs.rotation)}) activeUploads=\{oldTbs.initiating.activeCounter.get()} (self included)
								  took \{STANDARD_DURATION.nanosToString(elapsed)} in total
								    \{STANDARD_DURATION.nanosToString(elapsedPinned)} pinned
								    \{STANDARD_DURATION.nanosToString(elapsedSyncing)} syncing""", List.of(DEBUG));
						} finally {
							var oldTbs = tbs;

							oldTbs.initiating.onAvailableSync.register();

							// We're finished, so decrement the active count.
							// It is important that we do this only *after* registering to onAvailableSync to avoid
							// race conditions.
							oldTbs.initiating.activeCounter.getAndAdd(-1);

							// wait for the next to be available
							oldTbs.initiating.onAvailableSync.arriveAndAwaitAdvance();

							var newTbs = this.tbs.get();
							assert newTbs.completing == oldTbs.initiating;
							tbs = newTbs;

							if (++i != TBS_COUNT) {
								LOGGER.log(() -> STR."[Upload #\{id}] Ready for next upload", List.of(DEBUG));

								var n = newTbs.initiating.activeCounter.addAndGet(1);
								assert n > 0;
							}

							oldTbs.initiating.onAvailableSync.arriveAndDeregister();

							if (i == 1) {
								// The first upload has just completed. Now, spawn a thread to await the rotation
								// to the texture, then notify the caller that the upload has completed. At that
								// point, to an external observer, everything *has* completed. However, the changes
								// will still need to be propagated to the remaining two buffers internally.
								Thread.startVirtualThread(() -> {
									LOGGER.log(() -> STR."[Upload #\{id}] Waiting for first upload to rotate into active", List.of(DEBUG));
									try {
										if (newTbs.initiating.onAvailableSync.register() >= 0) {
											LOGGER.log(() -> STR."[Upload #\{id}] Awaiting", List.of(DEBUG));
											newTbs.initiating.onAvailableSync.arriveAndAwaitAdvance();
											newTbs.initiating.onAvailableSync.arriveAndDeregister();
										}
										LOGGER.log(() -> STR."[Upload #\{id}] Complete", List.of(DEBUG));
										completer.complete(null);
									} catch (Throwable e) {
										completer.completeExceptionally(e);
									}
								});
							}
						}
					}
				} finally {
					GL.glDeleteBuffer(bufId);
				}
			} catch (Throwable e) {
				completer.completeExceptionally(e);
			}
		});
		completer.join();
		LOGGER.log(() -> STR."[Upload #\{id}] Uploaded image to \{handle}", List.of(DEBUG));
	}

	@Override
	public void close() {
		this.textureArrayA.close();
		this.textureArrayB.close();
		this.textureArrayC.close();
	}

	public void dump() {
		this.textureArrayA.dump();
		this.textureArrayB.dump();
		this.textureArrayC.dump();
	}

	private static int tbBufIndex(byte rotation) {
		return rotation % TBS_COUNT;
	}

	private TextureArray getArray(byte rotation) {
		if (rotation < 0 || rotation > TBS_COUNT) throw new AssertionError();
		return switch (tbBufIndex(rotation)) {
			case 0 -> this.textureArrayA;
			case 1 -> this.textureArrayB;
			case 2 -> this.textureArrayC;
			default -> throw new AssertionError();
		};
	}

	private @NotNull TBS rotateTBS(TBS tbs, TBS nextTbs) {
		LOGGER.log(() -> STR."Rotating TBS:\n\t\{tbs}\n\t\{nextTbs}", List.of(DEBUG));
		long start = System.nanoTime();

		this.tbs.set(nextTbs);
//			tbs.completing.nextAvailable.countDown();
		// Two-step process. First announce that the next is available, then wait for other threads to submit their
		// jobs to it.
		// TODO: don't block the main thread. Instead, arrive and then check if others have arrived on each call to
		//  this method.
		nextTbs.completing.onAvailableSync.arriveAndAwaitAdvance();
		nextTbs.completing.onAvailableSync.arriveAndAwaitAdvance();
		nextTbs.completing.onAvailableSync.forceTermination();
		long elapsed = System.nanoTime() - start;
		LOGGER.log(() -> STR."Rotated TBS in \{STANDARD_DURATION.nanosToString(elapsed)}", List.of(DEBUG));
		return nextTbs;
	}

	/**
	 * @param rotation the rotation of the active texture for drawing from the zeroth.
	 */
	private record TBS(byte rotation, TBS1 completing, TBS1 initiating) {
		public TBS() {
			this((byte) 0, new TBS1(), new TBS1());
		}

		@Nullable
		public TBS pollNext() {
			if (this.completing.dirty.get() || this.initiating.dirty.get()) {
				if (this.completing.activeCounter.compareAndSet(0, Integer.MAX_VALUE)) {
					return new TBS((byte) ((byte) (this.rotation + 1) % 3), this.initiating, new TBS1());
				}
			}
			return null;
		}
	}

	/**
	 * <p>
	 * registering:
	 * MAX_VALUE + 1 -> MIN_VALUE
	 * => signals that we cannot add
	 * <p>
	 * completing:
	 * 1 - 1 -> 0
	 * <p>
	 * checking:
	 * compareAndSet(0, Integer.MAX_VALUE)
	 * => if true, all are complete, and no more will be added.
	 *
	 * @param dirty whether the buffer is being or has been modified
	 * @param activeCounter the number of active uploads to the buffer
	 * @param onAvailableSync triggered when the previous batch (another {@link TBS1}) is complete and the next batch
	 *                        has been made available.
	 */
	private record TBS1(AtomicBoolean dirty, AtomicInteger activeCounter, Phaser onAvailableSync) {
		public TBS1() {
			this(new AtomicBoolean(), new AtomicInteger(), new Phaser(1));
		}
	}
}
