package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.Canvas;
import dev.pfaff.altargui.internal.render.CanvasImpl;
import dev.pfaff.altargui.LoadableTex2d;
import dev.pfaff.altargui.MountContext;
import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.async.Task;
import dev.pfaff.altargui.render.objects.IRectRenderObject;
import dev.pfaff.altargui.render.objects.RectRenderObject;
import dev.pfaff.altargui.render.objects.TexturedRectRenderObject;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.texture.TextureHandle;
import dev.pfaff.altargui.types.Color;
import dev.pfaff.altargui.types.Tex2d;
import dev.pfaff.altargui.widgets.state.Painter;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.VarHandle;
import java.util.Objects;

import static dev.pfaff.altargui.async.Task.completedTask;
import static dev.pfaff.altargui.async.VirtualThreadTask.spawnVirtualTask;

/**
 * @param colorRgba the RGBA colorRgba of the widget
 * @param tex the texture of the widget
 */
public record RectWidget(int colorRgba, @Nullable LoadableTex2d tex) implements LeafWidget {
	public static final RectWidget TRANSPARENT = new RectWidget(0, null);
	public static final RectWidget GENERIC_HUD_BACKGROUND = colored(Color.fromArgb(0x90_505050));
	public static final RectWidget GENERIC_HEAVY_HUD_BACKGROUND = colored(Color.fromArgb(0xcc_505050));

	/**
	 * @deprecated treat this as private.
	 */
	@Deprecated
	public RectWidget {
	}

	/**
	 * @param color ARGB
	 */
	public static RectWidget colored(Color color) {
		if (color.alpha() == 0) return TRANSPARENT;
		return new RectWidget(color.rgba(), null);
	}

	public static RectWidget textured(LoadableTex2d tex) {
		return new RectWidget(Color.WHITE.rgba(), Objects.requireNonNull(tex));
	}

	public static RectWidget textured(LoadableTexture tex, Tex2d coords) {
		return textured(new LoadableTex2d(tex, coords));
	}

	public static RectWidget textured(LoadableTexture tex) {
		return textured(new LoadableTex2d(tex));
	}

	@Override
	public @NonNull WidgetInstance<RectWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public Color color() {
		return Color.fromRgba(colorRgba);
	}

	public RectWidget withColor(Color color) {
		return new RectWidget(color.rgba(), tex);
	}

	public RectWidget withTexture(LoadableTex2d tex) {
		return new RectWidget(colorRgba, tex);
	}

	private static final class State extends LeafWidget.AbstractState<RectWidget> implements Painter {
		private static final VarHandle VH_loaded;

		static {
			try {
				VH_loaded = MethodHandles.lookup().findVarHandle(State.class, "loaded", boolean.class);
			} catch (NoSuchFieldException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
		}

		private IRectRenderObject rect;
		private int color;
		private @Nullable LoadableTex2d tex;
		private boolean loaded = true;
		private @Nullable Task<@Nullable Void> loadingTask = null;
		private @Nullable TextureHandle texHandle;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, RectWidget widget, boolean init) {
			var assumeDirty = false;
			if (!(widget.tex != null
					? this.rect instanceof TexturedRectRenderObject
					: this.rect instanceof RectRenderObject)) {
				this.rect = widget.tex != null ? new TexturedRectRenderObject() : new RectRenderObject();
				assumeDirty = true;
			}
			if (!Objects.equals(widget.tex, this.tex)) {
				if (this.loadingTask != null) {
					this.loadingTask.close();
					this.loadingTask = null;
				}
				this.tex = widget.tex;
				this.texHandle = null;
				if (this.tex != null) {
					VH_loaded.setRelease(this, false);
				} else {
					VH_loaded.setRelease(this, true);
//					this.loaded = true;
//					if (rect instanceof TexturedRectRenderObject textured) {
//						// the white pixel texture should be locked on program 0 at 0, 0, 1, 1
////						textured.updateTexLayer(0);
////						textured.updateUV(0f, 0f, 0f, 0f);
//						// FIXME
//					}
				}
			}
			if (assumeDirty || widget.colorRgba != this.color) {
				var color = this.color = widget.colorRgba;
				if ((boolean) VH_loaded.getAcquire(this)) {
					this.rect.updateColor(Color.fromRgba(color));
				} else {
					this.rect.updateColor(Color.TRANSPARENT);
				}
			}
		}

		@Override
		protected void computeStencilLayout(StateContext context) {
			computeFillLayout(this);
			this.rect.updateSize(this.computedWidth(), this.computedHeight());
		}

		@Override
		protected void computeTraceLayout() {
			this.rect.updatePosition(this.allocatedX(), this.allocatedY());
		}

		@Override
		public void mount(MountContext context) {
			super.mount(context);
		}

		@Override
		public void paint(long time, Canvas canvas) {
			if (Color.fromRgba(this.color).alpha() == 0) return;

			var rect = this.rect;

			if (rect instanceof TexturedRectRenderObject) {
				if (this.loadingTask == null) {
					startLoadingTexture(canvas, rect);
				}
			}

			rect.uSpecUploadCanvas((CanvasImpl) canvas);
			//canvas.upload(rect);
		}

		private void startLoadingTexture(Canvas canvas, IRectRenderObject rect) {
			var tex = Objects.requireNonNull(this.tex);
			var coords = tex.coords();
			var allocTask = canvas.texAlloc().alloc(tex.tex());
			if (allocTask.isCompleted()) {
				var handle = allocTask.joinUninterruptibly();
				applyLoadedTexture(rect, handle, tex, coords);
				this.loadingTask = completedTask(null);
			} else {
				this.loadingTask = spawnVirtualTask(() -> {
					var handle = allocTask.join();
					applyLoadedTexture(rect, handle, tex, coords);
				});
			}
		}

		private void applyLoadedTexture(IRectRenderObject rect, TextureHandle handle, LoadableTex2d tex, Tex2d coords) {
			if (this.tex == tex && this.rect == rect) {
				this.texHandle = handle;

				var textured = (TexturedRectRenderObject) rect;
				textured.updateTex(handle, coords);
				VH_loaded.setRelease(this, true);
				rect.updateColor(Color.fromRgba(this.color));
			}
		}
	}
}
