package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.async.Task;
import dev.pfaff.altargui.async.VirtualThreadTask;
import dev.pfaff.altargui.widgets.state.StatefulWidgetInstance;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

/**
 * @param task    The task to listen to. Tracked by identity.
 * @param builder The builder function. Tracked by identity, low overhead of change.
 */
public record TaskBuilder<T>(Task<T> task, Builder<T> builder) implements Widget {
	public TaskBuilder(Task<T> task, Function<T, @NonNull Widget> successBuilder, Function<Throwable, @NonNull Widget> failureBuilder) {
		this(task, successBuilder, failureBuilder, Widgets.EMPTY);
	}

	public TaskBuilder(Task<T> task, Function<T, @NonNull Widget> successBuilder, Function<Throwable, @NonNull Widget> failureBuilder
			, @NonNull Widget childWhileLoading) {
		this(task, new SplitBuilder<>(successBuilder, failureBuilder, childWhileLoading));
	}

	@Override
	public @NonNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State<>(args);
	}

	private static final class State<T> extends StatefulWidgetInstance<TaskBuilder<T>> {
		private static final AltResult NULL_RESULT = new AltResult(null);

		private final AtomicInteger generation = new AtomicInteger();
		private Task<T> currentTask;
		@Nullable
		private Task<Void> listenerTask;
		private Builder<T> builder;
		private @Nullable Object result;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected Widget build() {
			var result = this.result;
			Throwable ex;
			boolean c = result != null;
			if (!c) {
				ex = null;
			} else if (result instanceof State.AltResult failed) {
				ex = failed.ex();
				result = null;
			} else {
				ex = null;
			}
			return Objects.requireNonNull(builder.accept((T) result, ex, c), "builder returned null");
		}

		private void setFuture(Task<T> task) {
			int expectedGeneration = generation.addAndGet(1);
			currentTask = task;
			result = null;
			listenerTask = VirtualThreadTask.spawnVirtualTask(() -> {
				@Nullable T value;
				@Nullable Throwable throwable;
				try {
					value = task.join();
					throwable = null;
				} catch (Throwable t) {
					value = null;
					throwable = t;
				}
				// abort if generations do not match, meaning that the task was replaced.
				if (generation.get() != expectedGeneration) return;
				//noinspection UnreachableCode,ConstantValue
				result = throwable == null ? value == null ? NULL_RESULT : value : new AltResult(throwable);
				markNeedsRebuild();
			});
		}

		@Override
		public void configure(StateContext stateContext, TaskBuilder<T> widget, boolean init) {
			var dirty = init;
			if (!widget.builder.equals(builder)) {
				dirty = true;
				builder = widget.builder;
			}
			var fut = widget.task;
			if (fut != currentTask) {
				dirty = true;
				setFuture(fut);
			}
			if (dirty) markNeedsRebuild();
		}

		private record AltResult(@Nullable Throwable ex) {}
	}

	@FunctionalInterface
	public interface Builder<T> {
		@NonNull Widget accept(T value, @Nullable Throwable throwable, boolean complete);
	}

	private record SplitBuilder<T>(Function<T, @NonNull Widget> successBuilder,
									 Function<Throwable, @NonNull Widget> failureBuilder,
									 @NonNull Widget loadingWidget) implements Builder<T> {
		public SplitBuilder {
			Objects.requireNonNull(successBuilder);
			Objects.requireNonNull(failureBuilder);
			Objects.requireNonNull(loadingWidget);
		}

		@Override
		public @NonNull Widget accept(T t, @Nullable Throwable e, boolean c) {
			if (!c) return loadingWidget;
			if (e != null) return Objects.requireNonNull(failureBuilder.apply(e), "failureBuilder returned null");
			return Objects.requireNonNull(successBuilder.apply(t), "successBuilder returned null");
		}
	}
}
